package org.atea.nlptools.koreromaoriproxy;

import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.annotation.MultipartConfig;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.Part;

import com.google.gson.FieldNamingPolicy;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.atea.nlptools.koreromaoriproxy.models.AudioFilePart;
import org.atea.nlptools.koreromaoriproxy.models.TranscriptionResult;
import org.atea.nlptools.koreromaoriproxy.models.TranscriptionResultDto;
import org.atea.nlptools.koreromaoriproxy.models.TranscriptionResultMetadata;

@MultipartConfig
public class TestServlet extends HttpServlet
{
	private static final long serialVersionUID = 1L;

	private static final Logger logger = LogManager.getLogger(TranscriptionServlet.class);
	private static final Type transcriptionListType = new TypeToken<List<TranscriptionResult>>(){}.getType();

	private final Gson jsonSerialiser;

	public TestServlet()
	{
		jsonSerialiser = new GsonBuilder().setFieldNamingPolicy(FieldNamingPolicy.LOWER_CASE_WITH_UNDERSCORES).create();
	}

	@Override
	public void init() 
	{
		logger.info("TestServlet Initialised!");
	}

	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
		throws IOException
	{
		response.getWriter().println("You are using the test endpoint.");
	}
	
	@Override
	protected void doPost(HttpServletRequest request, HttpServletResponse response)
			throws ServletException, IOException
	{
		try	
		{
			logger.trace("POST request received.");

			response.setContentType("application/json; charset=UTF-8");
			PrintWriter writer = response.getWriter();
	
			// Attempt to get the key of every audio file part
			String audioFileKeysParameter = request.getParameter("audioFileKeys");
			if (audioFileKeysParameter == null)
			{
				response.sendError(400, "Form data was incorrect: missing audioFileKeys part.");
				logger.trace("Request rejected due: Form data was missing the audioFileKeys part.");
				return;
			}
			String[] audioFileKeys = audioFileKeysParameter.split("\\|");
	
			// Create our own representation of each part
			List<AudioFilePart> audioFileParts = new ArrayList<AudioFilePart>(audioFileKeys.length);
			for (String audioFileKey : audioFileKeys)
			{
				// Attempt to get the part that this key specifies should be present
				Part p = request.getPart(audioFileKey);
				if (p == null) {
					response.sendError(400, "Form data was incorrect: a part key listed in the audioFileKeys was not found.");
					logger.trace("Request rejected due: Form data was missing a client-specified part.");
					return;
				}

				AudioFilePart audioFilePart = AudioFilePart.fromPart(p);
				audioFileParts.add(audioFilePart);
			}

			// Transcribe each part
            for (AudioFilePart p : audioFileParts) {
                p.setTranscriptionResult(generateMockTranscription());
            }

			// Transform the Tuhituhi response into our own format
			List<TranscriptionResultDto> responses = new ArrayList<TranscriptionResultDto>(audioFileParts.size());
			for (AudioFilePart result : audioFileParts) {
				responses.add(TranscriptionResultDto.FromTranscriptionResult(result.getTranscriptionResult(), result.fileName));
			}

			// Send our result to the client
			String json = jsonSerialiser.toJson(responses, transcriptionListType);
			writer.append(json);
		}
		catch (Exception ex)
		{
			response.sendError(500, "Failed to process the request.");
			logger.error("Failed to complete API call", ex);
		}
	}

    private TranscriptionResult generateMockTranscription()
    {
        ArrayList<TranscriptionResultMetadata> metadata = new ArrayList<TranscriptionResultMetadata>(7);
        metadata.add(new TranscriptionResultMetadata('ŋ', 0.5, 0));
        metadata.add(new TranscriptionResultMetadata('a', 0.5, 0.2));
        metadata.add(new TranscriptionResultMetadata(' ', 0.5, 0.4));
        metadata.add(new TranscriptionResultMetadata('t', 0.5, 0.6));
        metadata.add(new TranscriptionResultMetadata('a', 0.5, 0.8));
        metadata.add(new TranscriptionResultMetadata('m', 0.5, 1));
        metadata.add(new TranscriptionResultMetadata('a', 0.5, 1.2));

        return new TranscriptionResult("Generated a mock transcription", metadata, "1.0-fake", true, "nga tama");
    }
}
