package org.atea.nlptools.koreromaoriproxy.services;

import java.io.IOException;
import java.io.InputStream;
import java.net.SocketTimeoutException;
import java.util.LinkedList;
import java.util.List;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.atea.nlptools.koreromaoriproxy.exceptions.ReoTuhituhiException;
import org.atea.nlptools.koreromaoriproxy.models.AudioFilePart;
import org.atea.nlptools.koreromaoriproxy.models.TranscriptionResult;
import org.atea.nlptools.koreromaoriproxy.services.HttpRequestService.HttpRequestException;

/**
 * Functions to interact with the Reo Tuhituhi API.
 */
public class ReoTuhituhiApiService
{
    private static final Logger logger = LogManager.getLogger(ReoTuhituhiApiService.class);

    private final Gson jsonSerialiser;
    private final String apiEndpoint;
    private final String apiKey;

    public ReoTuhituhiApiService(Gson jsonSerialiser, String apiEndpoint, String apiKey)
    {
        this.jsonSerialiser = jsonSerialiser;
        this.apiEndpoint = apiEndpoint;
        this.apiKey = apiKey;
    }

    /**
     * Queries the Reo Tuhituhi API to transcribe the given audio files.
     * 
     * @param audioFileParts The audio files to retrieve a transcription for.
     * @return A list of {@link TranscriptionResult} objects.
     * @throws HttpRequestException When the API call fails.
     * @throws JsonSyntaxException When the result cannot be parsed.
     */
    public List<AudioFilePart> getTranscriptions(Iterable<AudioFilePart> audioFileParts)
        throws HttpRequestException, IOException
    {
        List<AudioFilePart> apiResults = new LinkedList<AudioFilePart>();
        
        for (AudioFilePart part : audioFileParts)
        {
            TranscriptionResult res = null;
            try
            {
                res = getTranscription(part.dataStream);
            }
            catch (ReoTuhituhiException rtex)
            {
                res = new TranscriptionResult("The Reo Tuhituhi API didn't respond successfully.");
            }
            catch (JsonSyntaxException jsex)
            {
                res = new TranscriptionResult("That's on me! I failed to parse the response from the Reo Tuhituhi API.");
            }
            catch (HttpRequestException hrex)
            {
                if (hrex.getCause().getClass() == SocketTimeoutException.class) {
                    res = new TranscriptionResult("The Reo Tuhituhi API timed out.");
                }
                else {
                    throw hrex;
                }
            }
            
            part.setTranscriptionResult(res);
            apiResults.add(part);
        }

        return apiResults;
    }

    /**
     * Calls the Reo Tuhituhi API to transcribe a wave audio file.
     * 
     * @param audioStream The wave audio stream.
     * @return A {@link TranscriptionResult} object.
     * @throws HttpRequestException Thrown when the API call fails.
     * @throws JsonSyntaxException Thrown when the result cannot be parsed.
     * @throws ReoTuhituhiException Thrown when the Reo Tuhituhi API returns an invalid response.
     * @throws IOException Thrown at any time.
     */
    public TranscriptionResult getTranscription(InputStream audioStream)
        throws HttpRequestException, JsonSyntaxException, ReoTuhituhiException, IOException
    {
        logger.trace("Performing request to tuhituhi API. Audio bytes available to pass through: " + audioStream.available());

        HttpRequestService request = HttpRequestService
            .post(apiEndpoint)
            .connectTimeout(2000)
            .readTimeout(5000)
            .authorization("Basic " + apiKey)
            .send(audioStream);

        // Check that the request returned an OK status
        if (!request.ok())
        {
            logger.error
            (
                "The Reo Tuhituhi API returned a non-OK status code {} with message {}",
                request.code(),
                request.message()
            );

            throw new ReoTuhituhiException(request.code(), request.message(), "API returned a non-OK status code");
        }

        // Check that the content type is valid
        if (!request.contentType().equals("application/json"))
        {
            logger.error
            (
                "The Reo Tuhituhi API returned an invalid content type of {}. Provided content was {}",
                request.contentType(),
                request.body()
            );

            throw new ReoTuhituhiException(request.code(), request.message(), "API specified an invalid content type: " + request.contentType());
        }

        logger.trace("Tuhituhi API returned a successful result. Beginning deserialisation...");

        TranscriptionResult res = jsonSerialiser.fromJson(request.body(), TranscriptionResult.class);

        logger.trace("Tuhituhi response successfully deserialised");

        return res;
    }
}
